/* 
   The samba-bugs bug tracking system
   smtp utility functions

   Copyright (C) Andrew Tridgell 1997

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "jitterbug.h"

#define SMTP_PORT 25

static char *errorstr = NULL;

char *smtp_error(void)
{
	return errorstr;
}

static int smtp_open(char *dest)
{
	struct in_addr addr;
	int type = SOCK_STREAM;
	struct sockaddr_in sock_out;
	int res;

	/* create a socket to write to */
	res = socket(PF_INET, type, 0);
	if (res == -1) {
		return -1;
	}

	addr.s_addr = inet_addr(dest);
  
	sock_out.sin_addr = addr;
	sock_out.sin_port = htons(SMTP_PORT);
	sock_out.sin_family = PF_INET;

	if (connect(res,(struct sockaddr *)&sock_out,sizeof(sock_out))) {
		close(res);
		return -1;
	}

	return res;
}


int smtp_write(int fd, char *fmt, ...)
{
	int len;
	char *p;
	char buf[1000];
	va_list ap;  

	va_start(ap, fmt);
	
	vsprintf(buf, fmt, ap);

	len = strlen(buf);
	p = buf;
	while (len) {
		int l = write(fd, p, len);
		if (l <= 0) return 0;
		len -= l;
		p += l;
	}
	return 1;
}


static int smtp_waitfor(int fd, char *code)
{
	int matched=0;
	char c;

	while (read(fd,&c,1) == 1) {
		if (c == '\n') {
			matched = 1;
			continue;
		}
		if (matched) {
			if (code[matched-1] == '?' || code[matched-1] == c)
				matched++;
			else
				matched = 0;
		}
		if (matched == strlen(code)+1) return 1;
	}

	return 0;
}

void smtp_write_data(int fd, char *p)
{
	while (*p) {
		smtp_write(fd,"%c",*p);
		if (p[0] == '\n' && p[1] == '.') {
			smtp_write(fd,".");
		}
		p++;
	}
}


int smtp_start_mail(char *from, char *to, char *cc, char *bcc, char *subject, int len)
{
	char *p, *tok;
	int fd = smtp_open(SMTP_ADDR);
	if (fd == -1) {
		errorstr = "smtp_open failed";
		return -1;
	}

	smtp_write(fd,"EHLO localhost\r\n");
	if (!smtp_waitfor(fd, "2?0 ")) {
		errorstr = "EHLO failed";
		close(fd);
		return -1;
	}

	if (strchr(from,'\n') || strchr(from,'\r')) {
		errorstr = "bad from address";
		close(fd);
		return -1;
	}

	smtp_write(fd,"MAIL FROM: <%s> SIZE=%d\r\n", 
		   extract_address(from));


	if (!smtp_waitfor(fd, "250 ")) {
		errorstr = "MAIL FROM rejected";
		close(fd);
		return -1;
	}
	
	p = strdup(to);
	for (tok=strtok(p," ,\t"); tok; tok=strtok(NULL," ,\t")) {
		smtp_write(fd,"RCPT TO: <%s>\r\n", user_address(tok));
		if (!smtp_waitfor(fd, "250 ")) {
			errorstr = "RCPT TO rejected";
			close(fd);
			return -1;
		}
	}
	free(p);
	if (cc && *cc) {
		p = strdup(cc);
		for (tok=strtok(p," ,\t"); tok; tok=strtok(NULL," ,\t")) {
			smtp_write(fd,"RCPT TO: <%s>\r\n", user_address(tok));
			if (!smtp_waitfor(fd, "250 ")) {
				errorstr = "RCPT TO rejected";
				close(fd);
				return -1;
			}
		}
		free(p);
	}
	if (bcc && *bcc) {
		p = strdup(bcc);
		for (tok=strtok(p," ,\t"); tok; tok=strtok(NULL," ,\t")) {
			smtp_write(fd,"RCPT TO: <%s>\r\n", user_address(tok));
			if (!smtp_waitfor(fd, "250 ")) {
				errorstr = "RCPT TO rejected";
				close(fd);
				return -1;
			}
		}
		free(p);
	}
	smtp_write(fd,"DATA\r\n");
	if (!smtp_waitfor(fd, "354 ")) {
		errorstr = "DATA rejected";
		close(fd);
		return -1;
	}
	if (subject) {
		smtp_write(fd,"From: %s\r\n", from);
		smtp_write(fd,"To: %s\r\n", to);
		smtp_write(fd,"Subject: %s\r\n", subject);
		if (cc && *cc)
			smtp_write(fd,"CC: %s\r\n", cc);
	}

	return fd;
}

int smtp_end_mail(int fd)
{
	smtp_write(fd,"\r\n.\r\n");
	if (!smtp_waitfor(fd, "250 ")) {
		errorstr = "end of data rejected";
		close(fd);
		return -1;
	}
	smtp_write(fd,"QUIT\r\n");
	if (!smtp_waitfor(fd, "221 ")) {
		errorstr = "QUIT rejected";
		close(fd);
		return -1;
	}
	close(fd);
	return 0;
}

